@@ -1,10 +1,11 @@
 from pyg_base import as_list, is_primitive, get_cache, dictattr, Dict, tree_getitem, getargspec, getargs, getcallargs, \
-                    eq, loop, ulist, tree_repr, wrapper, logger, is_strs, is_date, is_num, list_instances, is_ts, dictable
+                    eq, loop, ulist, tree_repr, wrapper, logger, is_strs, is_date, is_num, list_instances, is_ts, dictable, dictattr
                     
 from pyg_cell._dag import get_DAG, get_GAD, add_edge, del_edge, topological_sort 
-
+from pyg_encoders import as_writer, npy_write
 from copy import copy
 import datetime
+from functools import partial
 
 _data = 'data'
 _output = 'output'
@@ -56,7 +57,7 @@ def _cell_item(value, key = None, key_before_data = False):
             return value
     output = cell_output(value)
     if len(output) == 1:
-        if key is None:
+        if key is None: 
             return value[output[0]]
         else:
             if key_before_data:
@@ -102,6 +103,7 @@ def cell_outputs(value):
     return _cell_outputs(value)
     
 
+
 def cell_clear(value):
     """
     cell_clear clears a cell of its output so that it contains only the essentil stuff to do its calculations.
@@ -142,6 +144,22 @@ def cell_clear(value):
     else:
         return value
 
+def cell_repr(value):
+    if isinstance(value, cell):
+        return value._repr()
+    elif isinstance(value, (tuple, list)):
+        return type(value)([cell_repr(v) for v in value])
+    elif isinstance(value, dict):
+        try:
+            return type(value)(**{k : cell_repr(v) for k, v in value.items()})
+        except TypeError:
+            if isinstance(value, dictable):
+                return type(value)(**{str(k): cell_repr(v) for k, v in value.items()})
+            else:
+                return type(value)({k: cell_repr(v) for k, v in value.items()})
+    else:
+        return value
+
 
 def cell_item(value, key = None, key_before_data = False):
     """
@@ -495,6 +513,10 @@ class cell(dictattr):
         if _function not in kwargs:
             kwargs[_function] = None
         super(cell, self).__init__(**kwargs)
+        outputs = self._output
+        overlapping = {key : value for key, value in self._inputs.items() if isinstance(value, cell) and key in outputs}
+        if len(overlapping):
+            raise ValueError(f'cannot have a cell as an input and at the same time an output: {overlapping}, as it will be over-written on calculation, breaking graph structure')
         self.pull()
 
 
@@ -526,6 +548,7 @@ class cell(dictattr):
                     return True
         return False
     
+
     def save(self):
         """
         Saves the cell for persistency. For an in-memory cell, this is implemented by storing cell._address in the GRAPH
@@ -534,15 +557,48 @@ class cell(dictattr):
         -------
         cell
             self, saved.
+            
+            
+        Example:
+        -------
+        >>> from pyg import * 
+        >>> a = pd.Series([1,2,3], drange(2))
+        >>> b = 3
+        >>> c = cell(add_, a = a, b = b, name = 'james', surname = 'cohen', db = 'c:/test/%name/%surname.pickle')()
 
         """
+        writer = self._writer()
+        if writer is not None:
+            w = as_writer(writer)[0]
+            doc = {k: self[k] for k in self._output if k in self}
+            if len(doc)>0:
+                pk = self._pk
+                keys = {k : self[k] for k in pk if k in self}
+                if len(keys) < len(pk):
+                    logger.info(f'cannot save cell as need {pk} but only have {keys}')
+                else:
+                    doc.update(keys)
+                    _ = w(doc)        
         return self
 
 
     @property
     def _pk(self):
-        return ulist(sorted(as_list(self.get(_pk, self.get(_db)))))
-            
+        """
+        res = 'c:/%a/%b.parquet'
+        """
+        res = self.get(_pk, self.get(_db))
+        if isinstance(res, str) and '%' in res:
+            keys = [key.split('.')[0].split('/')[0].split('\\')[0] for key in res.split('%')[1:]]
+        else:
+            keys = as_list(res)
+        return ulist(sorted(as_list(keys)))
+    
+    def _writer(self):
+        db = self.get(_db)
+        if isinstance(db, str) and '%' in db:
+            return db
+        return None
 
         
     def load(self, mode = 0):
@@ -901,8 +957,14 @@ class cell(dictattr):
     def copy(self):
         return copy(self)
 
+    
+    def _repr(self):
+        pk = self._pk 
+        return tree_repr(type(self)({k:self.get(k) for k in pk + ['db', 'function'] + self._output}))
+
+
     def __repr__(self):
-        return '%s\n%s'%(self.__class__.__name__,tree_repr({k : cell_clear(v) for k, v in self.items()}))
+        return '%s\n%s'%(self.__class__.__name__,tree_repr({k : cell_repr(v) for k, v in self.items()}))
 
     def pull(self):
         """
