@@ -1,4 +1,4 @@
-from pyg_base import dt, eq, is_date, ulist, is_primitive, is_str, is_strs, as_list, get_cache, Dict, dictable, list_instances
+from pyg_base import loop, dt, eq, is_date, ulist, is_str, is_strs, as_list, get_cache, Dict, dictable, list_instances
 from pyg_encoders import cell_root, root_path, pd_read_parquet, pickle_load, pd_read_csv, dictable_decode
 from pyg_npy import pd_read_npy
 from pyg_cell._types import _get_mode, _get_qq, DBS, QQ, CLSS
@@ -15,6 +15,15 @@ def is_partial(db):
 def not_partial(db):
     return not isinstance(db, partial)
 
+@loop(list, tuple, dict)
+def impartial(v):
+    if isinstance(v, partial):
+        args, keywords = impartial((v.args, v.keywords))
+        return v.func(*args, **keywords)
+    else:
+        return v
+       
+
 _readers = dict(parquet = pd_read_parquet, 
                 pickle = pickle_load, 
                 npy = pd_read_npy, 
@@ -35,6 +44,7 @@ _function = 'function'
 _doc = 'doc'
 _writer = 'writer'
 
+
 __all__ = ['db_load', 'db_save', 'db_cell', 'cell_push', 'cell_pull', 'get_cell', 'load_cell', 'get_data', 'load_data']
 
 
@@ -57,6 +67,7 @@ def is_pairs(pairs):
     """
     return isinstance(pairs, tuple) and min([isinstance(item, tuple) and len(item) == 2 for item in pairs], default= False)
 
+    
 
 class db_cell_func(cell_func):
     """
@@ -195,7 +206,7 @@ def _is_primitive(value):
     elif isinstance(value, dict):
         return min([_is_primitive(v) for v in value.values()], default = True)
     else:
-        return value is None or isinstance(value, (type, bool, str, int, float, datetime.datetime, datetime.date, partial))
+        return value is None or isinstance(value, (type, bool, str, int, float, datetime.datetime, datetime.date, partial, cell))
 
 class db_cell(cell):
     """
@@ -263,14 +274,15 @@ class db_cell(cell):
             super(db_cell, self).__init__(function = function, output = output, **kwargs)
 
     _func = db_cell_func
-    
+    _shared_resources = ['engine', 'session', 'connection']
+
     @property
     def _function(self):
         db = self.get(_db)
         if not_partial(db):
             function = self.function if isinstance(self.function, cell_func) else cell_func(self.function)
         else:
-            keywords = self.db.keywords
+            keywords = self._keywords(self.db.keywords)
             env = {}
             for k in ['server', 'url']:
                 if k in keywords:
@@ -297,7 +309,40 @@ class db_cell(cell):
         db = self.get(_db)
         if not_partial(db):
             return super(db_cell, self)._pk
-        return self.db.keywords.get(_pk)        
+        return self.db.keywords.get(_pk)
+    
+    def _keywords(self, keywords):
+        keywords = {k: self.get(v, v) if is_str(v) and k in self._shared_resources and not is_str(self.get(v, v)) else v for k, v in keywords.items()}
+        keywords = {k: cell_item(v()) if isinstance(v, cell) else v for k, v in keywords.items()}
+        keywords = impartial(keywords)
+        return keywords   
+    
+    def _db(self, **kwargs):
+        """
+        equivalent to self.db() but allows the user to 
+        1) pass into partial function cells 
+        2) pass pointers to complicated elements within the cell
+    
+        Example: suppport for partials inside the db parameter:
+        --------
+        >>> from pyg import * 
+        >>> self = db_cell(add_, a = 1, b = 2, db = partial(sql_table, server = 'DESKTOP-LU5C5QF', db = partial(lambda db: db, db = 'test2'), 
+                                                            table = 'test7', 
+                                                            schema = 'cta', pk = 'key', create = True, doc = True), key = 'test')()
+
+    
+        """
+        value = cell_item(self.get(_db))
+        if is_partial(value):
+            keywords = self._keywords(value.keywords)
+            args = value.args
+            func = value.func
+            func = cell_item(func()) if isinstance(func, cell) else func
+            keywords.update(kwargs)
+            return func(*args, **keywords)
+        else:
+            return self.db
+            
 
     @property
     def _address(self):
@@ -305,8 +350,8 @@ class db_cell(cell):
         :Example:
         ----------
         >>> from pyg import *
-        >>> self = db_cell(db = partial(mongo_table, 'test', 'test', pk = 'key'), key = 1)
-        >>> db = self.db()
+        >>> self = db_cell(db = partial(sql_table, server = 'DESKTOP-LU5C5QF', db = 'test', table = 'test', schema = 'cta', pk = 'key', create = True), key = 1)
+        >>> self._db()
         >>> self._address
         >>> self._reference()
         >>> self.get('key')
@@ -338,7 +383,8 @@ class db_cell(cell):
         if not_partial(self.get(_db)): 
             return super(db_cell, self)._clear()
         else:
-            return self[[_db, _function, _updated] + self.db()._pk] if _updated in self else self[[_db, _function] + self.db()._pk]
+            pk = self._db()._pk
+            return self[[_db, _function, _updated] + pk] if _updated in self else self[[_db, _function] + pk]
 
     def save(self):
         db = self.get(_db)
@@ -346,7 +392,7 @@ class db_cell(cell):
             return super(db_cell, self).save()
         address = self._address
         doc = (self - _deleted)
-        db = db()
+        db = self._db()
         missing = ulist(db._pk) - self.keys()
         if len(missing):
             self._logger.warning('WARN: document not saved as some keys are missing %s'%missing)
@@ -360,12 +406,41 @@ class db_cell(cell):
             doc[i] = updated.get(i)
         get_GRAPH()[address] = doc
         return doc
-                
+    
+    def _needed(self, go):
+        """
+        boolean function, returns True if needs to load the cell from the database
+        """
+        spec = self.fullargspec
+        if spec is None:
+            return True
+        args = spec.args
+        missing = [arg for arg in args if self.get(arg) is None]
+        if len(missing):
+            return True
+        if go != 0 or self.run(): ## we will run anyway
+            return False
+        else:
+            return True
+   
         
-    def load(self, mode = 0):
+    def load(self, mode = 0, go = 0):
         """
         loads a document from the database and updates various keys.
         
+        Parameters:
+        ----------
+        mode: 
+            
+            0: load from the graph. If missing from graph, load from db
+            -1: do not load, clear from graph
+            'if needed': 
+                if the cell is going to be re-evaluated anyway, don't load from the database. 
+                if cell.run() is true, then we load it only if input and output overlap.
+                
+                
+        
+        
         Example:Persistency:
         -------------
         Since we want to avoid hitting the database, there is a singleton GRAPH, a dict, storing the cells by their address.
@@ -392,10 +467,9 @@ class db_cell(cell):
         
         Example:Merge of cached cell and calling cell:
         ----------------------------------------
-        Once we load from memory (either MongoDB or GRAPH), we tree_update the cached cell with the new values in the current cell.
-        This means that it is next to impossible to actually *delete* keys. If you want to delete keys in a cell/cells in the database, you need to:
-        
-        >>> del db.inc(filters)['key.subkey']
+        Once we load from memory (either MongoDB or GRAPH), we update the cached cell with the output values in the current cell.
+        You can choose to load the cell and import it in its entirety if you set load(mode = 'load and import')
+    
 
         :Parameters:
         ----------
@@ -415,17 +489,7 @@ class db_cell(cell):
         """
         if not_partial(self.get(_db)):
             return super(db_cell, self).load(mode = mode)
-        # if isinstance(mode, (list, tuple)):
-        #     if len(mode) == 0:
-        #         return self.load()
-        #     if len(mode) == 1:
-        #         return self.load(mode[0])
-        #     if len(mode) == 2 and mode[0] == -1:
-        #         res = self.load(-1)
-        #         return res.load(mode[-1])
-        #     else:
-        #         raise ValueError('mode can only be of the form [], [mode] or [-1, mode]')
-        db = self.db(mode = 'w')
+        db = self._db(mode = 'w')
         pk = ulist(db._pk)
         missing = pk - self.keys()
         if len(missing):
@@ -443,6 +507,8 @@ class db_cell(cell):
                 doc = _load_asof(db, kwargs, deleted = mode)
                 graph[doc._address] = doc
             else:
+                if mode == 'if needed' and not self._needed(go):
+                    return self
                 try:
                     doc = _load_asof(table = db, kwargs = kwargs, deleted = False, qq = None)
                     graph[doc._address] = doc
@@ -456,17 +522,22 @@ class db_cell(cell):
             saved = graph[address] 
             self_updated = self.get(_updated)
             saved_updated = saved.get(_updated)
-            if self_updated is None or (saved_updated is not None and saved_updated > self_updated):
-                excluded_keys = (self /  None).keys() - self._output - _updated
+            if mode == 'load and update' or self.function is None:
+                if self_updated is None or (saved_updated is not None and saved_updated > self_updated):
+                    excluded_keys = (self /  None).keys() - self._output - _updated
+                else:
+                    excluded_keys = (self /  None).keys()
+                if is_date(mode):
+                    excluded_keys += [_id]
+                update = (saved / None) - excluded_keys
             else:
-                excluded_keys = (self /  None).keys()
-            if is_strs(mode):
-                excluded_keys += as_list(mode)
-            elif is_date(mode):
-                excluded_keys += [_id]
-            update = (saved / None) - excluded_keys
+                if self_updated is None or (saved_updated is not None and saved_updated > self_updated):
+                    update = {key : saved[key] for key in saved._output + [_updated] if key in saved and saved[key] is not None}
+                else:
+                    update = {}
             updated_inputs = [k for k, v in self._inputs.items() if k in saved and v is not None and not isinstance(v, cell) and not eq(saved[k], v)]
-            self.update(update)
+            if update:
+                self.update(update)
             if len(updated_inputs):
                 self[_updated] = None
         return self        
@@ -571,7 +642,7 @@ class db_cell(cell):
         if db is None:
             return super(db_cell, self).bind(**bind)
         else:
-            kw = self.db.keywords
+            kw = self._keywords(self.db.keywords)
             for k in bind: # we want to be able to override tables/db/url
                 if k in ['db', 'table', 'url']:
                     kw[k] = bind.pop(k)
@@ -582,6 +653,21 @@ class db_cell(cell):
             res = res(**bind)
             res[_db] = db
             return self + res
+        
+class calculating_cell(db_cell):
+    """
+    This is a cell that when saved in the database, its calculated values are not saved.
+    """
+    def save(self):
+        output_keys = self._output
+        output = {key : self[key] for key in output_keys if key in self}
+        pre_save = db_cell(self - output_keys)
+        res = type(self)(pre_save.save())
+        res.update(output)
+        address = res._address
+        get_GRAPH()[address] = res
+        return res
+        
 
 
 def cell_push(nodes = None, exc = None, go = 1):
@@ -981,3 +1067,5 @@ def load_data(table = None, db = None, url = None, schema = None, server = None,
     """    
     return cell_item(load_cell(table, db = db, url = url, schema = schema, server = server, deleted = deleted, doc = doc, **kwargs), key = 'data')
 
+
+
